static int x_fd;
static int tty_fd;

static Ring ttyoutbuf;

static void
open_connection(void)
{
	int err;
	conn = xcb_connect(NULL, NULL);
	err = xcb_connection_has_error(conn);
	if (err)
		panic(connection_error_message(err));
	x_fd = xcb_get_file_descriptor(conn);
}

static void
init(void)
{
	// Not shown. Opens the X connection and forks the shell process,
	// among other things.
}

static void
handle_event(xcb_generic_event_t *e)
{
	// Not shown. May write into ttyoutbuf.
}

static void
poll_x(void)
{
	xcb_generic_event_t *e;
	for (;;) {
		e = xcb_poll_for_event(conn);
		if (e == NULL) {
			if (xcb_connection_has_error(conn))
				panic("X IO error");
			return;
		}
		if (e->response_type)
			handle_event(e);
		else
			handle_error((xcb_generic_error_t *)e);
		free(e);
	}
}

static void
loop(void)
{
	struct pollfd fds[] = {
		{ .fd = x_fd, .events = POLLIN },
		{ .fd = tty_fd, .events = POLLIN },
	};
	int n, eof;
	Ring inbuf;
	ringinit(&inbuf, malloc(512), 512); // TODO leak ok?
	for (;;) {
		if (ringsize(&ttyoutbuf) > 0)
			fds[1].events |= POLLOUT;
		else
			fds[1].events &= ~POLLOUT;
		n = poll(fds, 2, -1);
		if (n == -1) {
			if (errno == EAGAIN)
				continue;
			if (errno == EINTR) {
				// TODO what goes here?
				continue;
			}
			panic("poll: %m");
		}
		if (fds[0].revents & POLLIN)
			poll_x();
		if (fds[1].revents & POLLIN) {
			ringread(&inbuf, tty_fd, &eof);
			if (eof)
				exit(0);
			if (ringsize(&inbuf) > 0) {
				parse(&inbuf);
				if (is_mapped && is_visible)
					draw();
			}
		}
		if (fds[1].revents & POLLOUT)
			ringwrite(&ttyoutbuf, tty_fd);
	}
}

int
main(void)
{
	init();
	loop();
	return 0;
}
